import disnake
from disnake import TextInputStyle
from disnake.ext import commands
import pandas as pd
from typing import List
from webull_options.webull_options import WebullOptions
import os
from tabulate import tabulate

from discord_.bot_menus.pagination import AlertMenus


from dotenv import load_dotenv
load_dotenv()
theta_em = "<a:_:1047010360554704977>"
gamma_em = "<a:_:1190024505700134973>"
delta_em = "<a:_:1044647851466182736>"
vega_em = "<a:_:1044647903009976372>"
rho_em = "<a:_:1044647745702596648>"
ticker_em = "<:_:1190025407815221270>"
volume_em = "🔊"
oi_em = "👥"
oi_change_em = "🔀"
iv_em="💉"
expiry_em="🗓️"
strike_em="🎳"
call_put_em="↕️"
change_ratio_em="➗"
start_over_em="<:_:1190026332248232126>"
open_em="<a:_:1044647404693106728>"
high_em="<a:_:1047010108930019340>"
low_em="<a:_:1044658254137008148>"
close_em="<a:_:1044647558133334126>"
underlying_price_em="💲"
sdk = WebullOptions(access_token=os.environ.get('ACCESS_TOKEN'),osv=os.environ.get('OSV'), did=os.environ.get('DID'))

class FilterMenu(disnake.Embed):
    def __init__(self):
        super().__init__( 
            title=f"Options Filter Menu",
            description=f"```py\nMenu Buttons:```\n\n> {theta_em} - Theta\n> {delta_em} - Delta\n> {gamma_em} - Gamma\n> {vega_em} - Vega\n> {rho_em} - Rho\n\n> {iv_em} - IV\n> {volume_em} - Vol\n> {oi_em} - OI\n> {oi_change_em} - OI Change\n> {change_ratio_em} - Change%\n\n> {open_em} - Open\n> {high_em} - High\n> {low_em} - Low\n> {close_em} - Close\n> {underlying_price_em} - Underlying Price\n\n> {ticker_em} - Ticker\n> {strike_em} - Strike\n> {call_put_em} - Direction\n> {expiry_em} - Expiry\n\n> {start_over_em} - Clear"
        )


        self.set_footer(text=f"Implemented by FUDSTOP")
        self.add_field(name=f"Info:", value=f"> Select up to 5 attributes to query from the drop-down menu below.")




class QueryMenu(disnake.ui.Select):
    def __init__(self):
        # Use provided options if they are not None, otherwise default to an empty list


        super().__init__(custom_id='results', row=2, min_values=1, max_values=5, options= [ 
            disnake.SelectOption(label='underlying_symbol', description='Filter options based on the ticker name.',emoji=ticker_em),
            disnake.SelectOption(label='theta', description=f'Filter options by min / max theta values.', emoji=theta_em),
            disnake.SelectOption(label='gamma', description='Filter options by min/max gamma values.', emoji=gamma_em),
            disnake.SelectOption(label='vega', description='Filter options by min/max vega values.', emoji=vega_em),
            disnake.SelectOption(label='imp_vol', description='Filter options by min/max IV values.', emoji=delta_em),
            disnake.SelectOption(label='volume', description='Filter options by min/max volume values.', emoji=delta_em),
            disnake.SelectOption(label='open_interest', description='Filter options by min/max open interest values.', emoji=oi_em),
            disnake.SelectOption(label='open_int_change', description='Filter options by min/max oi change values.', emoji=oi_change_em),
            disnake.SelectOption(label='open', description='Filter options by min/max open price values.', emoji=open_em),
            disnake.SelectOption(label='high', description='Filter options by min/max high price values.', emoji=high_em),
            disnake.SelectOption(label='low', description='Filter options by min/max low price values.', emoji=low_em),
            disnake.SelectOption(label='close', description='Filter options by min/max latest price values.', emoji=close_em),
            disnake.SelectOption(label='change_ratio', description='Filter options by min/max change% values.', emoji=change_ratio_em),
            disnake.SelectOption(label='strike_price', description='Filter options by min/max strike price values.', emoji=strike_em),
            disnake.SelectOption(label='expire_date', description='Filter options by expiration.', emoji=expiry_em),
            disnake.SelectOption(label='call_put', description='Filter options by calls or puts.', emoji=call_put_em),
            disnake.SelectOption(label='underlying_price', description='Filter options by min/max underlying price values.', emoji=underlying_price_em),


        ])

    async def callback(self, inter: disnake.AppCmdInter):
        text_inputs =[]
        for value in self._selected_values:
            
            components = [
            disnake.ui.TextInput(
                label=f"Filter by {value}",
                placeholder=f"The minimum {value} you wish the option to have.",
                custom_id=f"{value}",
                style=TextInputStyle.short
            )]

            text_inputs.append(components)

        await inter.response.send_modal(FilterModal(components=text_inputs))




class FilterView(disnake.ui.View):
    def __init__(self):
        super().__init__(timeout=None)
        self.add_item(QueryMenu())
       

       




# Subclassing the modal.
class FilterModal(disnake.ui.Modal):
    def __init__(self, components):
        self.components = components
        super().__init__(title="Create Tag", components=self.components)

    def chunk_string(self, string, size):
        """Yield successive size-sized chunks from string."""
        for i in range(0, len(string), size):
            yield string[i:i + size]

    async def callback(self, inter: disnake.ModalInteraction):
        conn = await sdk.db_manager.get_connection()
        await inter.response.defer()
        # Operator mapping for different fields
        operator_mapping = {
            'expire_date': '=',
            'underlying_symbol': '=',
            'imp_vol': '>=',
            'open_int_change': '>=',
            'vega': '>=',
            'gamma': '>=',
            'delta': '>=',
            'volume': '>=',
            'open_interest': '>=',
            'strike_price': '='
            # Add other fields and their corresponding operators here
        }

        # Fields that need to be wrapped in single quotes
        quote_fields = ['underlying_symbol']

        # Prepare the base query and conditions
        query = "SELECT underlying_symbol, strike_price, call_put, expire_date FROM options WHERE "
        conditions = []

        for key, value in inter.text_values.items():
            # Choose operator based on the field, defaulting to '='
            operator = operator_mapping.get(key, '=')

            # Wrap value in quotes if the field is in quote_fields, else directly insert the value
            if key in quote_fields:
                formatted_value = f"'{value}'"
            else:
                formatted_value = value

            conditions.append(f"{key} {operator} {formatted_value}")

        # Combine conditions and finalize the query
        query += " AND ".join(conditions) + " LIMIT 25;"
        print(query)
        results = await conn.fetch(query)
        df = pd.DataFrame(results)
        print(df.columns)


        df = df.rename(columns={'underlying_symbol': 'sym', 'strike_price':'strike', 'call_put': 'cp', 'expire_date': 'exp'})
        table = tabulate(df, headers=['sym', 'strike', 'cp', 'expiry', 'theta'], tablefmt='fancy', showindex=False)
        # Break apart data into chunks of 4000 characters
        chunks = self.chunk_string(table, 4000)
        embeds=[]
        # Create and send embeds for each chunk
        for chunk in chunks:
            embed = disnake.Embed(title="Option Filter Results", description=f"```py\n{chunk}```")
            embeds.append(embed)
        embed.add_field(name=f"Query:", value=f"```py\n{query}```")
        await inter.edit_original_message(embed=embeds[0],view=AlertMenus(embeds).add_item(QueryMenu()))  # Use send or edit_original_message as appropriate






import os
from dotenv import load_dotenv
load_dotenv()

from openai import OpenAI
from openai.resources.beta.assistants.assistants import Assistants
from openai.resources.beta.threads.threads import Threads
from openai.resources.beta.threads.runs.runs import Runs
from openai.resources.beta.threads.messages.messages import Messages

client = OpenAI(api_key=os.environ.get('YOUR_OPENAI_KEY'))



def create_assistant(name:str, instructions:str, tools:str, model:str='gpt-4-1106-preview'):
    """
    Creates an OpenAI assistant using the API.

    Arguments:


    >>> Name: the name of the assistant.
    >>> Instructions: the instructions for the assistant.
    >>> Tools: The tools to give the assistant. e.g. code-interpreter, retrieval
    >>> Model: the model to use. default = gpt-4-1106-preview
    """
    assistant = client.beta.assistants.create(
        name=name,
        instructions=f"{instructions}",
        tools=[{"type": f"{tools}"}],
        model=model
    )
    id = assistant.id
    model = assistant.model
    instructions = assistant.instructions
    return assistant


def create_thread():
    thread = client.beta.threads.create()

    return thread.id


def create_message(thread, role:str, content:str):
    """
    Creates a message within a thread between user / assistant.

    Arguments:

    >>> thread_id: the ID of the thread
    >>> role: whether the message is from the user or the assistant
    >>> content: the content of the interaction
    """

    message = client.beta.threads.messages.create(
        thread_id=thread.id,
        role=role,
        content=content
    )


    return message

def run_thread(thread_id):
    run = client.beta.threads.runs.create(
    thread_id='thread_xogYNUjD5psm38lbxYZxGmeI',
    assistant_id='asst_mzbMbtc8p0jjUomVYPaGARPB',
    instructions="Please address the user as Jane Doe. The user has a premium account."
    )

    return run.id


def check_run(thread_id, run_id):
    run = client.beta.threads.runs.retrieve(
    thread_id=thread_id,
    run_id=run_id
    )

    return run

assistant_id='asst_mzbMbtc8p0jjUomVYPaGARPB'
thread_id= 'thread_xogYNUjD5psm38lbxYZxGmeI'
run_id = run_thread(thread_id)

thread_check = check_run(thread_id, run_id)

print(thread_check)


import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from dotenv import load_dotenv
load_dotenv()
import asyncpg
from asyncpg.exceptions import UniqueViolationError, ForeignKeyViolationError, DataError
import asyncpg
import uuid
import os
from typing import List
from datetime import datetime
from asyncpg import create_pool
from datetime import date
import json
import asyncio
import pandas as pd
import numpy as np
from asyncpg.exceptions import UniqueViolationError, ForeignKeyViolationError
import pytz
from asyncio import Lock
from datetime import timezone
lock = Lock()
password = os.environ.get('DB_PASSWORD') # Default password if not found in environment variables
db_config_fudstop = {
    "host": os.environ.get('DB_HOST'), # Default to this IP if 'DB_HOST' not found in environment variables
    "port": int(os.environ.get('DB_PORT')), # Default to 5432 if 'DB_PORT' not found
    "user": 'chuck', # Default to 'postgres' if 'DB_USER' not found
    "password": password, # Use the password from environment variable or default
    "database": os.environ.get('DB_NAME') # Database name for the new jawless database
}

db_options_chat_config =  { 
    "host": os.environ.get('DB_HOST'), # Default to this IP if 'DB_HOST' not found in environment variables
    "port": int(os.environ.get('DB_PORT')), # Default to 5432 if 'DB_PORT' not found
    "user": os.environ.get('DB_USER'), # Default to 'postgres' if 'DB_USER' not found
    "password": password, # Use the password from environment variable or default
    "database": os.environ.get('DB_NAME') # Database name for the new jawless database
}
class DbManager:
    def __init__(self, host, port, user, password, database):
        self.conn = None
        self.pool = None
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.connection_string = f"postgresql://{user}:{password}@{host}:{port}/{database}"
        print(self.connection_string)
        print(self.user)
        self.chat_memory = []  # In-memory list to store chat messages

    async def connect(self):
        self.pool = await create_pool(
            host='localhost', user='chuck', database='markets', port=5432, password='fud', min_size=1, max_size=10
        )

        return self.pool
    
    async def get_ticker_id_by_symbol(self, option_symbol):
        async with self.pool.acquire() as connection:
            query = "SELECT ticker_id FROM webull_options WHERE symbol = $1"
            ticker_id = await connection.fetchval(query, option_symbol)
            
            # Check if ticker_id was found
            if ticker_id is not None:
                return (True, ticker_id)
            else:
                return (False, None)

    async def insert_specials(self, ticker, strike, call_put, expiry, buy_pct, neutral_pct, sell_pct,
                                official_open, last_price, day_vwap, buy_volume, neutral_volume, sell_volume,
                                last_volume, avg_price, volume_percent_total, total_trades, moneyness,
                                price_diff, price_change_pct, volume_to_price_ratio, option_id):
        conn = await self.connect()
        query = """
        INSERT INTO volanal (ticker, strike, call_put, expiry, buy_pct, neutral_pct, sell_pct,
                                official_open, last_price, day_vwap, buy_volume, neutral_volume, sell_volume,
                                last_volume, avg_price, volume_percent_total, total_trades, moneyness,
                                price_diff, price_change_pct, volume_to_price_ratio, option_id)
        VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22)
        """
        await conn.execute(query, ticker, strike, call_put, expiry, buy_pct, neutral_pct, sell_pct,
                            official_open, last_price, day_vwap, buy_volume, neutral_volume, sell_volume,
                            last_volume, avg_price, volume_percent_total, total_trades, moneyness,
                            price_diff, price_change_pct, volume_to_price_ratio, option_id)

    async def options_chat_connect(self):

        self.pool = await create_pool(
            **db_options_chat_config, min_size=1, max_size=20
        )

        return self.pool
    async def insert_chat_history(self, sender: str, message: str, session_id: uuid.UUID, connection_string: str):
        conn = await asyncpg.connect(connection_string)
        query = """INSERT INTO chat_history (sender, message, session_id, created_at) VALUES ($1, $2, $3, $4);"""
        await conn.execute(query, sender, message, session_id, datetime.now())
   

    async def batch_insert_ticker_snaps(self, records: list):
        async with self.pool.acquire() as conn:
            insert_query = '''
            INSERT INTO ticker_snapshots (
                underlying_ticker, day_close, day_open, day_low, day_high, day_volume,
                day_vwap, prev_close, prev_high, prev_low, prev_open, prev_volume,
                prev_vwap, change_percent, change, min_high, min_low, min_open,
                min_close, min_volume, min_vwap, min_accumulated_volume, last_trade_size,
                last_trade_price, last_trade_conditions, last_trade_exchange, last_trade_timestamp,
                ask, ask_size, bid, bid_size, quote_timestamp
            ) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16,
                    $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28, $29, $30,
                    $31, $32, $33);
            '''
            await conn.executemany(insert_query, records)
  
    async def execute_query_and_fetch_one(self, query, parameters=None):
        async with self.pool.acquire() as conn:
            if parameters:
                row = await conn.fetchrow(query, *parameters)
            else:
                row = await conn.fetchrow(query)
            return row
    async def check_if_exists(self, link, table_name):
        query = f"SELECT COUNT(*) FROM {table_name} WHERE link = ?"
        count = await self.execute_query_and_fetch_one(query, (link,))
        return count[0] > 0


    async def insert_td9_state(self, ticker: str, status: str, timespan: str):
        async with self.pool.acquire() as connection:
            try:
                await connection.execute(
                    """
                    INSERT INTO td9_states (ticker, status, timespan) VALUES ($1, $2, $3);
                    """,
                    ticker, status, timespan
                )
            except UniqueViolationError:
                print(f"SKIPPING {ticker} {status} {timespan}")
    async def fetch(self, query):
        async with self.pool.acquire() as conn:
            records = await conn.fetch(query)
            return records
    async def clear_old_td9_states(self):
        async with self.pool.acquire() as connection:
            await connection.execute(
                """
                DELETE FROM td9_states WHERE timestamp <= (current_timestamp - interval '4 minutes');
                """
            )

    async def process_response(self, response):
        async with self.pool.acquire() as conn:
            for result in response['results']:
                # Convert Unix Msec timestamp to datetime in Eastern timezone
                timestamp = result['t'] / 1000  # Convert milliseconds to seconds
                timestamp_eastern = timestamp.astimezone(timezone('US/Eastern'))
                
                # Map single-letter columns to their full names
                result['open'] = result.pop('o')
                result['high'] = result.pop('h')
                result['low'] = result.pop('l')
                result['close'] = result.pop('c')
                result['volume'] = result.pop('v')

                # Insert data into the table
                await conn.execute('''
                    INSERT INTO stock_data (ticker, adjusted, query_count, request_id, results_count,
                                            status, open, high, low, n, close, otc, t, volume, vw)
                    VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15)
                ''', (response['ticker'], response['adjusted'], response['queryCount'], response['request_id'],
                      response['resultsCount'], response['status'], result['open'], result['high'], result['low'],
                      result['n'], result['close'], result['otc'], timestamp_eastern, result['volume'], result['vw']))

    async def disconnect(self):
        await self.pool.close()
    async def safe_batch_insert_dataframe(self, df, table_name, unique_columns, retries=3):
        for i in range(retries):
            try:
                await self.batch_insert_dataframe(df, table_name, unique_columns)
                break  # Successful, break the loop
            except Exception as e:
                print(f"An error occurred: {e}")
                if i < retries - 1:
                    print("Retrying...")
                else:
                    print("Max retries reached. Raising the exception.")
                    raise

    async def insert_symbol(self, symbol):
        async with self.pool.acquire() as connection:
            insert_query = '''
            INSERT INTO symbols (symbol)
            VALUES ($1)
            ON CONFLICT (symbol) DO NOTHING;
            '''
            await connection.execute(insert_query, symbol)

    async def insert_sma_data(self, records):
        # records is a list of tuples, where each tuple corresponds to a row to be inserted
        async with self.pool.acquire() as conn:
            await conn.executemany(
                """
                INSERT INTO sma_data (sma_value, timestamp, sma_length, ticker, timespan, something)
                VALUES ($1, $2, $3, $4, $5, $6)
                ON CONFLICT (ticker, timestamp, timespan) DO UPDATE
                SET sma_value = EXCLUDED.sma_value,
                    sma_length = EXCLUDED.sma_length,
                    something = EXCLUDED.something
                """,
                records
            )
    async def insert_options_snapshot(self, data_dict):
        async with self.pool.acquire() as connection:
            insert_query = '''
            INSERT INTO options_data (name, option_symbol, ticker, strike, "call put", expiry, "underlying price",
            "change", "change percent", "early change", "early change_percent", "change to break_even", "break even price",
            open, high, low, close, "previous close", volume, oi, iv, delta, gamma, theta, vega, "trade size", "trade price",
            "trade exchange", "trade conditions", "trade timestamp", ask, "ask size", "ask exchange", bid, "bid size", "bid exchange")
            VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22,
            $23, $24, $25, $26, $27, $28, $29, $30, $31, $32, $33, $34, $35, $36)
            '''
            await connection.execute(insert_query, *data_dict.values())
    async def insert_tickers(self, ticker_data):
        async with self.pool.acquire() as connection:
            async with connection.transaction():
                # Prepare the SQL INSERT query
                insert_query = '''
                INSERT INTO tickers (ticker_symbol, todays_change, todays_change_perc)
                VALUES ($1, $2, $3)
                ON CONFLICT (ticker_symbol) DO UPDATE 
                SET todays_change = EXCLUDED.todays_change,
                    todays_change_perc = EXCLUDED.todays_change_perc;
                '''
                
                # Execute the query in a batch
                await connection.executemany(insert_query, [(d['ticker'], d['todaysChange'], d['todaysChangePerc']) for d in ticker_data])

    async def table_exists(self, table_name):
        query = f"SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = '{table_name}');"
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                exists = await conn.fetchval(query)
        return exists
    


    async def create_dynamic_view(self, table_name, features_dict):
        feature_expressions = ", ".join([
            f"({expression}) as {feature_name}"
            for feature_name, expression in features_dict.items()
        ])
        create_view_query = f"""
        CREATE OR REPLACE VIEW {table_name}_with_features AS
        SELECT *, {feature_expressions}
        FROM {table_name};
        """

        async with self.pool.acquire() as connection:
            await connection.execute(create_view_query)


    async def insert_embedding(self, original_text, embedding):
        query = """
        INSERT INTO embeddings (original_text, embedding)
        VALUES ($1, $2);
        """
        async with self.pool.acquire() as connection:
            await connection.execute(query, original_text, embedding)

    async def get_embedding(self, original_text):
        query = """
        SELECT embedding FROM embeddings
        WHERE original_text = $1;
        """
        async with self.pool.acquire() as connection:
            return await connection.fetchval(query, original_text)

    async def create_table(self, df, table_name, unique_column):
        print("Connected to the database.")
        dtype_mapping = {
            'int64': 'INTEGER',
            'float64': 'FLOAT',
            'object': 'TEXT',
            'bool': 'BOOLEAN',
            'datetime64': 'TIMESTAMP',
            'datetime64[ns]': 'timestamp',
            'datetime64[ms]': 'timestamp',
            'datetime64[ns, US/Eastern]': 'TIMESTAMP WITH TIME ZONE'
        }
        print(f"DataFrame dtypes: {df.dtypes}")
        # Check for large integers and update dtype_mapping accordingly
        for col, dtype in zip(df.columns, df.dtypes):
            if dtype == 'int64':
                max_val = df[col].max()
                min_val = df[col].min()
                if max_val > 2**31 - 1 or min_val < -2**31:
                    dtype_mapping['int64'] = 'BIGINT'
        history_table_name = f"{table_name}_history"
        async with self.pool.acquire() as connection:

            table_exists = await connection.fetchval(f"SELECT to_regclass('{table_name}')")
            
            if table_exists is None:
                unique_constraint = f'UNIQUE ({unique_column})' if unique_column else ''
                create_query = f"""
                CREATE TABLE {table_name} (
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))},
                    "insertion_timestamp" TIMESTAMP,
                    {unique_constraint}
                )
                """
                print(f"Creating table with query: {create_query}")

                # Create the history table
                history_create_query = f"""
                CREATE TABLE IF NOT EXISTS {history_table_name} (
                    id serial PRIMARY KEY,
                    operation CHAR(1) NOT NULL,
                    changed_at TIMESTAMP NOT NULL DEFAULT current_timestamp,
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))}
                );
                """
                print(f"Creating history table with query: {history_create_query}")
                await connection.execute(history_create_query)
                try:
                    await connection.execute(create_query)
                    print(f"Table {table_name} created successfully.")
                except asyncpg.UniqueViolationError as e:
                    print(f"Unique violation error: {e}")
            else:
                print(f"Table {table_name} already exists.")
            
            # Create the trigger function
            trigger_function_query = f"""
            CREATE OR REPLACE FUNCTION save_to_{history_table_name}()
            RETURNS TRIGGER AS $$
            BEGIN
                INSERT INTO {history_table_name} (operation, changed_at, {', '.join(f'"{col}"' for col in df.columns)})
                VALUES (
                    CASE
                        WHEN (TG_OP = 'DELETE') THEN 'D'
                        WHEN (TG_OP = 'UPDATE') THEN 'U'
                        ELSE 'I'
                    END,
                    current_timestamp,
                    {', '.join('OLD.' + f'"{col}"' for col in df.columns)}
                );
                RETURN NEW;
            END;
            $$ LANGUAGE plpgsql;
            """
            await connection.execute(trigger_function_query)

            # Create the trigger
            trigger_query = f"""
            DROP TRIGGER IF EXISTS tr_{history_table_name} ON {table_name};
            CREATE TRIGGER tr_{history_table_name}
            AFTER UPDATE OR DELETE ON {table_name}
            FOR EACH ROW EXECUTE FUNCTION save_to_{history_table_name}();
            """
            await connection.execute(trigger_query)


            # Alter existing table to add any missing columns
            for col, dtype in zip(df.columns, df.dtypes):
                alter_query = f"""
                DO $$
                BEGIN
                    BEGIN
                        ALTER TABLE {table_name} ADD COLUMN "{col}" {dtype_mapping[str(dtype)]};
                    EXCEPTION
                        WHEN duplicate_column THEN
                        NULL;
                    END;
                END $$;
                """
                await connection.execute(alter_query)

    async def insert_dataframe(self, df, table_name, unique_columns):
        # Check if the table already exists
        if not await self.table_exists(table_name):
            await self.create_table(df, table_name, unique_columns)  # Assuming this function exists

        # Convert DataFrame to list of records
        records = df.to_records(index=False)
        data = list(records)

        # Get a connection from the connection pool
        async with self.pool.acquire() as connection:
            # Fetch the PostgreSQL table schema to get column types (Only once)
            column_types = await connection.fetch(f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'")
            column_type_dict = {item['column_name']: item['data_type'] for item in column_types}

            # Create a transaction
            async with connection.transaction():
                # Prepare the INSERT query
                insert_query = f"""
                INSERT INTO {table_name} ({', '.join(f'"{col}"' for col in df.columns)}) 
                VALUES ({', '.join('$' + str(i) for i in range(1, len(df.columns) + 1))})
                ON CONFLICT ({unique_columns})
                DO UPDATE SET {', '.join(f'"{col}" = excluded."{col}"' for col in df.columns)}
                """

                async def execute_query(record):
                    new_record = []
                    for col, val in zip(df.columns, record):
                        pg_type = column_type_dict.get(col, None)

                        # Conversion logic starts here
                        if val is None:
                            new_record.append(None)
                        elif pg_type == 'timestamp' and isinstance(val, np.datetime64):
                            print(f"Converting {val} to datetime")  # Debugging print statement
                            new_record.append(pd.Timestamp(val).to_pydatetime())
                        elif isinstance(val, np.datetime64) and np.isnat(val):
                            new_record.append(None)

                        elif pg_type == 'timestamp without time zone' and isinstance(val, np.datetime64):
                            new_record.append(pd.Timestamp(val).to_pydatetime() if pd.notnull(val) else None)
                        elif pg_type == 'timestamp with time zone' and isinstance(val, np.datetime64):
                            new_record.append(pd.Timestamp(val).to_pydatetime() if pd.notnull(val) else None)
                        elif pg_type in ['double precision', 'real']:
                            if isinstance(val, str):
                                new_record.append(float(val))
                            elif isinstance(val, pd.Timestamp):
                                # Convert to datetime in eastern timezone or whatever you need to do
                                val = val.tz_convert('US/Eastern')
                                new_record.append(val)
                            else:
                                new_record.append(float(val))
                        elif pg_type == 'integer' and not isinstance(val, int):
                            new_record.append(int(val))
                        else:
                            new_record.append(val)
                        # Conversion logic ends here


                    try:
                        await connection.execute(insert_query, *new_record)
                    except Exception as e:
                        print(f"An error occurred while inserting the record: {e}")
                        await connection.execute('ROLLBACK')
                        raise

                # Execute the query for each record concurrently
                await asyncio.gather(*(execute_query(record) for record in data))








    async def batch_insert_dataframe(self, df, table_name, unique_columns, batch_size=250):
        async with lock:
            if not await self.table_exists(table_name):
                await self.create_table(df, table_name, unique_columns)

            # Debug: Print DataFrame columns before modifications
            #print("Initial DataFrame columns:", df.columns.tolist())
            
            df = df.copy()
            df.dropna(inplace=True)
            df['insertion_timestamp'] = [datetime.now() for _ in range(len(df))]

            # Debug: Print DataFrame columns after modifications
            #print("Modified DataFrame columns:", df.columns.tolist())
            
            records = df.to_records(index=False)
            data = list(records)


            async with self.pool.acquire() as connection:
                column_types = await connection.fetch(
                    f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'"
                )
                type_mapping = {col: next((item['data_type'] for item in column_types if item['column_name'] == col), None) for col in df.columns}

                async with connection.transaction():
                    insert_query = f"""
                    INSERT INTO {table_name} ({', '.join(f'"{col}"' for col in df.columns)}) 
                    VALUES ({', '.join('$' + str(i) for i in range(1, len(df.columns) + 1))})
                    ON CONFLICT ({unique_columns})
                    DO UPDATE SET {', '.join(f'"{col}" = excluded."{col}"' for col in df.columns)}
                    """
            
                    batch_data = []
                    for record in data:
                        new_record = []
                        for col, val in zip(df.columns, record):
                   
                            pg_type = type_mapping[col]

                            if val is None:
                                new_record.append(None)
                            elif pg_type == 'timestamp' and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))

            
                            elif isinstance(val, datetime):
                                new_record.append(pd.Timestamp(val).to_pydatetime())
                            elif pg_type in ['timestamp', 'timestamp without time zone', 'timestamp with time zone'] and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))  # Modified line
                            elif pg_type in ['double precision', 'real'] and not isinstance(val, str):
                                new_record.append(float(val))
                            elif isinstance(val, np.int64):  # Add this line to handle numpy.int64
                                new_record.append(int(val))
                            elif pg_type == 'integer' and not isinstance(val, int):
                                new_record.append(int(val))
                            else:
                                new_record.append(val)
                    
                        batch_data.append(new_record)

                        if len(batch_data) == batch_size:
                            try:
                                
                             
                                await connection.executemany(insert_query, batch_data)
                                batch_data.clear()
                            except Exception as e:
                                print(f"An error occurred while inserting the record: {e}")
                                await connection.execute('ROLLBACK')
                                raise

                if batch_data:  # Don't forget the last batch
       
                    try:

                        await connection.executemany(insert_query, batch_data)
                    except Exception as e:
                        print(f"An error occurred while inserting the record: {e}")
                        await connection.execute('ROLLBACK')
                        raise
    async def save_to_history(self, df, main_table_name, history_table_name):
        # Assume the DataFrame `df` contains the records to be archived
        if not await self.table_exists(history_table_name):
            await self.create_table(df, history_table_name, None)

        df['archived_at'] = datetime.now()  # Add an 'archived_at' timestamp
        await self.batch_insert_dataframe(df, history_table_name, None)



import pandas as pd
from typing import Optional, List, Dict, Any
class DailyTreasury:
    def __init__(self, data):
        self.date = [i.get('Date') if i.get('Date') is not None else None for i in data]
        self.close_balance = [i.get('Close Balance') if i.get('Close Balance') is not None else None for i in data]
        self.open_balance = [i.get('Open Balance') if i.get('Open Balance') is not None else None for i in data]
        self.amount_change = [i.get('Amount Change') if i.get('Amount Change') is not None else None for i in data]
        self.percent_change = [i.get('Percent Change') if i.get('Percent Change') is not None else None for i in data]
        self.moving_avg = [i.get('Moving Avg') if i.get('Moving Avg') is not None else None for i in data]

        self.data_dict = {
                'Date': self.date,
                'Close Balance': self.close_balance,
                'Open Balance': self.open_balance,
                'Amount Change': self.amount_change,
                'Percent Change': self.percent_change,
                'Moving Avg': self.moving_avg
            }
        self.as_dataframe = pd.DataFrame(self.data_dict)
        


class EarningsCalendar:
    def __init__(self, data):
        self.date = [i['date'] if i['date'] is not None else None for i in data]
        self.hour = [i['hour'] if i['hour'] is not None else None for i in data]
        self.ticker = [i['ticker'] if i['ticker'] is not None else None for i in data]
        self.eps_est = [i['eps_est'] if i['eps_est'] is not None else None for i in data]
        self.eps_act  = [i['eps_act'] if i['eps_act'] is not None else None for i in data]
        self.revenue_est = [i['revenue_est'] if i['revenue_est'] is not None else None for i in data]
        self.revenue_act = [i['revenue_act'] if i['revenue_act'] is not None else None for i in data]
        self.year = [i['year'] if i['year'] is not None else None for i in data]
        self.quarter = [i['quarter'] if i['quarter'] is not None else None for i in data]
        self.mkt_cap = [i['mkt_cap'] if i['mkt_cap'] is not None else None for i in data]


        self.data_dict = {
            'date': self.date,
            'hour': self.hour,
            'ticker': self.ticker,
            'eps_est': self.eps_est,
            'eps_act': self.eps_act,
            'revenue_est': self.revenue_est,
            'revenue_act': self.revenue_act,
            'year': self.year,
            'quarter': self.quarter,
            'mkt_cap': self.mkt_cap
             }
        self.as_dataframe = pd.DataFrame(self.data_dict)


class FTD:
    def __init__(self, data):
        print(data)
        self.date = [i['Date'] if i['Date'] is not None else None for i in data]
        try:
            self.ticker = [i['Ticker'] if i['Ticker'] is not None else None for i in data]
        except KeyError:
            self.ticker = "N/A"
        self.price = [i['Price'] if i['Price'] is not None else None for i in data]
        self.amount_ftd = [i['Failure to Deliver'] if 'Failure to Deliver' in i else None for i in data]

        self.t35_date = [i['T+35 Date'] if i['T+35 Date'] is not None else None for i in data]
        self.dollar_cost = [i['Amount (FTD x $)'] if 'Amount (FTD x $)' in i else None for i in data]

        self.data_dict = {
            'Date': self.date,
            'Ticker': self.ticker,
            'Price': self.price,
            'Failure to Deliver': self.amount_ftd,
            'T+35 Date': self.t35_date,
            'Amount (FTD x $)': self.dollar_cost
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)

class House:
    def __init__(self, data):
        house = data['house']
        self.trans_date = [i['Transaction Date'] if i['Transaction Date'] in i else None for i in house]
        self.owner = [i['Owner'] if i['Owner'] in i else None for i in house]
        self.ticker = [i['Ticker'] if i['Ticker'] in i else None for i in house]
        self.asset_desc = [i['Asset Description'] if i['Asset Description'] in i else None for i in house]
        self.asset_type = [i['Asset Type'] if i['Asset Type'] in i else None for i in house]
        self.type = [i['Type'] if i['Type'] in i else None for i in house]
        self.amount = [i['Amount'] if i['Amount'] in i else None for i in house]
        self.rep = [i['Representative'] if i['Representative'] in i else None for i in house]
        self.link = [i['Link'] if i['Link'] in i else None for i in house]
        self.disclosure_date = [i['Disclosure Date'] if i['Disclosure Date'] in i else None for i in house]
        self.district=[i['District'] if i['District'] in i else None for i in house]
        self.capital_gains_over_200k = [i['Cap Gains Over 200USD'] if i['Cap Gains Over 200USD'] in i else None for i in house]


class Inflation:
    def __init__(self, data):
        self.shares = data.get('shares')
        self.value = data.get('value')
        self.shares_total = data.get('shares_total')

        self.data_dict = { 

            'Shares': self.shares,
            'Value': self.value,
            'Shares Total': self.shares_total
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)

class IPOs:
    def __init__(self, data):
        self.date = [i['Date'] if i['Date'] is not None else None for i in data]
        self.exchange = [i['Exchange'] if i['Exchange'] is not None else None for i in data]
        self.expected_price = [i['Expected Price'] if i['Expected Price'] is not None else None for i in data]
        self.mkt_cap = [i['Mkt Cap'] if i['Mkt Cap'] is not None else None for i in data]
        self.name = [i['Name'] if i['Name'] is not None else None for i in data]
        self.number_shares = [i['Number Shares'] if i['Number Shares'] is not None else None for i in data]
        self.status = [i['Status'] if i['Status'] is not None else None for i in data]
        self.symbol = [i['Symbol'] if i['Symbol'] is not None else None for i in data]


        self.data_dict = {
            'Date': self.date,
            'Exchange': self.exchange,
            'Expected Price': self.expected_price,
            'Mkt Cap': self.mkt_cap,
            'Name': self.name,
            'Number Shares': self.number_shares,
            'Status': self.status,
            'Symbol': self.symbol
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)
class JimCramer:
    def __init__(self, data):
        self.ticker = [i['Ticker'] if i['Ticker'] is not None else None for i in data]
        self.date = [i['Date'] if i['Date'] is not None else None for i in data]
        self.segment = [i['Segment'] if i['Segment'] is not None else None for i in data]
        self.call = [i['Call'] if i['Call'] is not None else None for i in data]
        self.price = [i['Price'] if i['Price'] is not None else None for i in data]

        self.data_dict = {
            'Ticker': self.ticker,
            'Date': self.date,
            'Segment': self.segment,
            'Call': self.call,
            'Price': self.price
        }


        self.as_dataframe = pd.DataFrame(self.data_dict)

class JoblessClaims:
    def __init__(self, data):
        self.date = [i['Date'] if i['Date'] is not None else None for i in data]
        self.number = [i['Number'] if i['Number'] is not None else None for i in data]
        self.percent_change = [i['Percent Change'] if i['Percent Change'] is not None else None for i in data]


        self.data_dict = {
            'Date': self.date,
            'Number': self.number,
            'Percent Change': self.percent_change
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)

class LowFloat:
    def __init__(self, data):
        self.Rank = [i['Rank'] if i['Rank'] is not None else None for i in data]
        self.ticker= [i['ticker'] if i['ticker'] is not None else None for i in data]
        self.company_name= [i['company_name'] if i['company_name'] is not None else None for i in data]
        self.exchange= [i['exchange'] if i['exchange'] is not None else None for i in data]
        self.previous_close= [i['previous_close'] if i['previous_close'] is not None else None for i in data]
        self.one_day_change= [i['one_day_change'] if i['one_day_change'] is not None else None for i in data]
        self.floating_shares= [i['floating_shares'] if i['floating_shares'] is not None else None for i in data]
        self.outstanding_shares= [i['outstanding_shares'] if i['outstanding_shares'] is not None else None for i in data]
        self.short_int= [i['short_int'] if i['short_int'] is not None else None for i in data]
        self.market_cap= [i['market_cap'] if i['market_cap'] is not None else None for i in data]
        self.industry= [i['industry'] if i['industry'] is not None else None for i in data]
        self.data_dict = {
            'Rank': self.Rank,
            'ticker': self.ticker,
            'company_name': self.company_name,
            'exchange': self.exchange,
            'previous_close': self.previous_close,
            'one_day_change': self.one_day_change,
            'floating_shares': self.floating_shares,
            'outstanding_shares': self.outstanding_shares,
            'short_int': self.short_int,
            'market_cap': self.market_cap,
            'industry': self.industry
            }
        
        self.as_dataframe = pd.DataFrame(self.data_dict)

class LatestInsiderTradingSummary:
    def __init__(self, data):
        self.ticker = [i['Ticker'] if i['Ticker'] is not None else None for i in data]
        self.amount = [i['Amount'] if i['Amount'] is not None else None for i in data]
        self.market_cap =[i['Market Cap'] if i['Market Cap'] is not None else None for i in data]
        self.percent_of_market_cap = [i['% of Mkt Cap'] if i['% of Mkt Cap'] is not None else None for i in data]
        self.data_dict = {
            'Ticker': self.ticker,
            'Amount': self.amount,
            'Market Cap': self.market_cap,
            '% of Mkt Cap': self.percent_of_market_cap
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)

class MarketNews:
    def __init__(self, data):
        self.Date = [i['Date'] if i['Date'] is not None else None for i in data]
        self.Title = [i['Title'] if i['Title'] is not None else None for i in data]
        self.Source = [i['Source'] if i['Source'] is not None else None for i in data]
        self.URL = [i['URL'] if i['URL'] is not None else None for i in data]
        self.Section = [i['Section'] if i['Section'] is not None else None for i in data]
        self.data_dict = {
            'Date': self.Date,
            'Title': self.Title,
            'Source': self.Source,
            'URL': self.URL,
            'Section': self.Section
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)

class MarketSummary:
    def __init__(self, data):
        self.snp500 = data.get('snp500')


class NewsSentiment:
    def __init__(self, data):
        self.Date = [i['Date'] if 'Date' in i else None for i in data]
        self.Title = [i['Title'] if 'Title' in i else None for i in data]
        self.Link = [i['Link'] if 'Link' in i else None for i in data]
        self.Sentiment = [i['Sentiment'] if 'Sentiment' in i else None for i in data]

class RetailSales:
    def __init__(self, data):
        self.date = [i['Date'] if i['Date'] is not None else None for i in data]
        self.amount = [i['Amount'] if i['Amount'] is not None else None for i in data]
        self.percent_change = [i['Percent Change'] if i['Percent Change'] is not None else None for i in data]
        self.monthly_avg_cases = [i['monthly_avg_cases'] if i['monthly_avg_cases'] is not None else None for i in data]

        self.data_dict = {
            'Date': self.date,
            'Amount': self.amount,
            'Percent Change': self.percent_change,
            'monthly_avg_cases': self.monthly_avg_cases
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)

class ReverseRepo:
    def __init__(self, data):
        self.date = [i['Date'] if i['Date'] is not None else None for i in data]
        self.amount = [i['Amount'] if i['Amount'] is not None else None for i in data]
        self.num_parties = [i['Num Parties'] if i['Num Parties'] is not None else None for i in data]
        self.average = [i['Average'] if i['Average'] is not None else None for i in data]
        self.moving_average = [i['Moving Average'] if i['Moving Average'] is not None else None for i in data]
        self.data_dict = {
            'Date': self.date,
            'Amount': self.amount,
            'Num Parties': self.num_parties,
            'Average': self.average,
            'Moving Average': self.moving_average
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)

class SECFillings:
    def __init__(self, data):
        self.Filling = [i['Filling'] if i['Filling'] is not None else None for i in data]
        self.Description = [i['Description'] if 'description' in i else None for i in data]
        self.FillingDate = [i['Filling Date'] if i['Filling Date'] is not None else None for i in data]
        self.report_url = [i['report_url'] if 'report_url' in i else None for i in data]
        self.filing_url = [i['filing_url'] if 'filing_url' in i else None for i in data]

        self.data_dict = {
            'Filling': self.Filling,
            'Description': self.Description,
            'Filling Date': self.FillingDate,
            'report_url': self.report_url,
            'filing_url': self.filing_url
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)
class Senate:

    def __init__(self, data):
        self.TransactionDate = [data.get('Transaction Date') for i in data]
        self.Owner = [data.get('Owner') if i['Owner'] is not None else None for i in data]
        self.Ticker = [i['Ticker'] if i['Ticker'] is not None else None for i in data]
        self.AssetDescription = [i['Asset Description'] if i['Asset Description'] is not None else None for i in data]
        self.AssetType = [data.get('Asset Type') if i['Asset Type'] is not None else None for i in data]
        self.Type = [data.get('Type') if i['Type'] is not None else None for i in data]
        self.Amount = [data.get('Amount') if i['Amount'] is not None else None for i in data]
        self.Senator = [data.get('Senator') if i['Senator'] is not None else None for i in data]
        self.Link = [data.get('Link') if i['Link'] is not None else None for i in data]
        self.DisclosureDate = [data.get('Disclosure Date') if i['Disclosure Date'] is not None else None for i in data]


class ShortInterest:
    def __init__(self, data):

        
        self.Rank = [i['Rank'] if i['Rank'] is not None else None for i in data]
        self.Ticker= [i['Ticker'] if i['Ticker'] is not None else None for i in data]
        self.Date=[i['Date'] if i['Date'] is not None else None for i in data]
        self.ShortInterest= [i['Short Interest'] if i['Short Interest'] is not None else None for i in data]
        self.AverageVolume= [i['Average Volume'] if i['Average Volume'] is not None else None for i in data]
        self.DaysToCover= [i['Days To Cover'] if i['Days To Cover'] is not None else None for i in data]
        self.FloatShort= [i['% Float Short'] if i['% Float Short'] is not None else None for i in data]

        self.data_dict = {
            'Rank': self.Rank,
            'Ticker': self.Ticker,
            'Date': self.Date,
            'Short Interest': self.ShortInterest,
            'Average Volume': self.AverageVolume,
            'Days To Cover': self.DaysToCover,
            '% Float Short': self.FloatShort
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)

class ShortVolume:
    def __init__(
        self,
        date: Optional[str] = None,
        shortVol: Optional[str] = None,
        shortExemptVol: Optional[str] = None,
        totalVol: Optional[str] = None,
        percentShorted: Optional[str] = None
    ):
        self.date = date
        self.shortVol = shortVol
        self.shortExemptVol = shortExemptVol
        self.totalVol = totalVol
        self.percentShorted = percentShorted

    def data_dict(self) -> Dict[str, Any]:
        return {
            'date': self.date,
            'shortVol': self.shortVol,
            'shortExemptVol': self.shortExemptVol,
            'totalVol': self.totalVol,
            'percentShorted': self.percentShorted
        }

    def data_dict(self) -> Dict[str, Any]:
        return {
            'date': self.date,
            'shortVol': self.shortVol,
            'shortExemptVol': self.shortExemptVol,
            'totalVol': self.totalVol,
            'percentShorted': self.percentShorted
        }

    @classmethod
    def from_dict(cls, data: List[Dict[str, Any]]) -> List['ShortVolume']:
        return [cls(
            date=item.get('Date'),
            shortVol=item.get('Short Vol'),
            shortExemptVol=item.get('Short Exempt Vol'),
            totalVol=item.get('Total Vol'),
            percentShorted=item.get('% Shorted')
        ) for item in data]

    def __iter__(self):
        return iter(self.data_dict().values())
class StockTwits:
    def __init__(self, data):
        self.rank = [i['rank'] if i['rank'] is not None else None for i in data]
        self.watchlist = [i['watchlist'] if i['watchlist'] is not None else None for i in data]
        self.date_updated = [i['date_updated'] if i['date_updated'] is not None else None for i in data]


        self.data_dict = { 

            'Rank': self.rank,
            'Watchlists': self.watchlist,
            'Date Updated': self.date_updated
        }


        self.as_dataframe = pd.DataFrame(self.data_dict)

class Subreddit:
    def __init__(self, data):
        self.Date = [i['Date'] if i['Date'] is not None else None for i in data]
        self.subreddit = [i['subreddit'] if i['subreddit'] is not None else None for i in data]
        self.Redditors = [i['Redditors'] if i['Redditors'] is not None else None for i in data]
        self.Active = [i['Active'] if i['Active'] is not None else None for i in data]
        self.percActive = [i['% Active'] if i['% Active'] is not None else None for i in data]
        self.percGrowth = [i['% Growth'] if i['% Growth'] is not None else None for i in data]
        self.percPriceChange = [i['% Price Change'] if i['% Price Change'] is not None else None for i in data]


        self.data_dict = {
            'Date': self.Date,
            'subreddit': self.subreddit,
            'Redditors': self.Redditors,
            'Active': self.Active,
            '% Active': self.percActive,
            '% Growth': self.percGrowth,
            '% Price Change': self.percPriceChange
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)
class TradingHalts:
    def __init__(self, data):
        self.halt_date = data.get('halt_date')
        self.halt_time = data.get('halt_time')
        self.ticker = data.get('ticker')
        self.exchange = data.get('exchange')
        self.reason = data.get('reason')
        self.resume_date = data.get('resume_date')
        self.resume_time = data.get('resume_time')

        self.data_dict = {
            'halt_date': self.halt_date,
            'halt_time': self.halt_time,
            'ticker': self.ticker,
            'exchange': self.exchange,
            'reason': self.reason,
            'resume_date': self.resume_date,
            'resume_time': self.resume_time
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)
class WSBMentions:
    def __init__(self, data):
        self.mentions = [i['mentions'] if i['mentions'] is not None else None for i in data]


class WSBOptions:
    def __init__(self, data):
        self.ticker = [i['Ticker'] if i['Ticker'] is not None else None for i in data]
        self.calls = [i['Calls'] if i['Calls'] is not None else None for i in data]
        self.puts = [i['Puts'] if i['Puts'] is not None else None for i in data]
        self.ratio = [i['Ratio'] if i['Ratio'] is not None else None for i in data]

        self.data_dict = {
            'Ticker': self.ticker,
            'Calls': self.calls,
            'Puts': self.puts,
            'Ratio': self.ratio
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)
class Insiders:
    def __init__(self, data):


        self.Ticker = [i['Ticker'] if i['Ticker'] is not None else None for i in data]
        self.Name = [i['Name'] if i['Name'] is not None else None for i in data]
        self.Relationship = [i['Relationship'] if i['Relationship'] is not None else None for i in data]
        self.Date = [i['Date'] if i['Date'] is not None else None for i in data]
        self.Transaction = [i['Transaction'] if i['Transaction'] is not None else None for i in data]
        self.Cost = [i['Cost'] if i['Cost'] is not None else None for i in data]
        self.Shares = [i['Shares'] if i['Shares'] is not None else None for i in data]
        self.Value = [i['Value ($)'] if i['Value ($)'] is not None else None for i in data]
        self.SharesTotal= [i['#Shares Total'] if i['#Shares Total'] is not None else None for i in data]
        self.DateFilled= [i['Date Filled'] if i['Date Filled'] is not None else None for i in data]



        self.data_dict = {
            'Ticker': self.Ticker,
            'Name': self.Name,
            'Relationship': self.Relationship,
            'Date': self.Date,
            'Transaction': self.Transaction,
            'Cost': self.Cost,
            'Shares': self.Shares,
            'Value ($)': self.Value,
            '#Shares Total': self.SharesTotal,
            'Date Filled': self.DateFilled
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)


import os
from dotenv import load_dotenv
import re
load_dotenv()
import disnake
from disnake import TextInput,TextInputStyle
import pandas as pd
from datetime import datetime
from tabulate import tabulate
from .webull_options import WebullOptions 


options = WebullOptions(os.environ.get('WEBULL_OPTIONS'))


class WebullModal(disnake.ui.Modal):
    def __init__(self, selected_columns, ticker:str):
        self.connection_string = os.environ.get('WEBULL_OPTIONS')
        
        self.ticker = ticker.upper()
        self.selected_columns = selected_columns
        # Placeholder text for specific fields
        placeholders = {
            'call_put': "Enter 'call' or 'put'",
            'expiry_date': "Enter a date (e.g., '2023-12-31')",
            'open_interest': "Enter a numerical threshold (e.g., '> 1000')",
            'open_interest_change': "Enter a change value (e.g., '< -50')",
            'strike_price': "Enter a strike price (e.g., '= 300')",
            # You can continue to add more fields with specific placeholders if needed.
            'volume': "Enter a volume amount (e.g., '>= 50000')",
            'bid_price': "Enter a bid price threshold (e.g., '< 5.00')",
            'ask_price': "Enter an ask price threshold (e.g., '> 5.00')",
            'theta': "Enter a theta value threshold",
            'gamma': "Enter a gamma value threshold",
            # Add any additional fields relevant to your options data and what users may want to query.
        }
        # Default placeholder for fields not in the dictionary
        default_placeholder = "Enter condition (e.g., '> 100')"

        components = []
        for column in selected_columns:  # Create a TextInput for each selected column
            components.append(
                disnake.ui.TextInput(
                    label=f"{column}",
                    placeholder=placeholders.get(column, default_placeholder),
                    custom_id=f"condition_{column}",
                    style=TextInputStyle.short,
                )
            )
        # Make sure to pass the components to the super().__init__
        super().__init__(title="Query Options Database", components=components)
        
    async def callback(self, inter: disnake.ModalInteraction):
            await options.connect()  # make sure you have a connection method that sets up your pool
            conditions = []
            for column in self.selected_columns:
            
                
                
                user_input = inter.text_values[f'condition_{column}'].strip()

                if user_input:
                    # Directly include user_input in the condition string
                    # Check if the column is 'expiry_date' and wrap the value in single quotes
                    if column == 'expiry_date':
                        match = re.match(r"([<>]=?|=?)\s*(\d{4}-\d{2}-\d{2})$", user_input)
                        operator, number = match.groups()
                        print(operator)
                        # For 'expiry_date', you expect the input in the format 'YYYY-MM-DD'
                        conditions.append(f"{column} {operator} '{user_input}'")
                   

                    if column == 'call_put':
                        # For 'expiry_date', you expect the input in the format 'YYYY-MM-DD'
                        conditions.append(f"{column} = '{user_input}'")


                    if column == 'strike_price':
                        # This regex matches operators followed by an integer or decimal number for strike_price
                        match = re.match(r"([<>]=?|=?)\s*(\d+(?:\.\d+)?)$", user_input)
                        if match:
                            operator, number = match.groups()
                            # Construct the condition string for strike_price
                            condition = f"{column} {operator} {number}"
                            conditions.append(condition)
                        else:
                            # If the input doesn't match the pattern, send an error message
                            await inter.response.send_message(f"Invalid format for strike price: {user_input}. Expected format: '<operator> <number>'")
                            return
                        
                                                    
                    match = None

                    if column == 'strike_price':
                        # Matches numeric conditions for strike_price
                        match = re.match(r"([<>]=?|=?)\s*(\d+(?:\.\d+)?)$", user_input)
                    elif column == 'expiry_date':
                        # Matches date conditions for expiry_date
                        match = re.match(r"([<>]=?|=?)\s*(\d{4}-\d{2}-\d{2})$", user_input)
                    elif column == 'open_interest':
                        # Matches numeric conditions for open_interest
                        match = re.match(r"([<>]=?|=?)\s*(\d+)$", user_input)
                    elif column == 'volume':
                        # Matches numeric conditions for volume
                        match = re.match(r"([<>]=?|=?)\s*(\d+)$", user_input)
                    elif column == 'bid_price':
                        # Matches decimal conditions for bid_price
                        match = re.match(r"([<>]=?|=?)\s*(\d+(?:\.\d+)?)$", user_input)
                    elif column == 'ask_price':
                        # Matches decimal conditions for ask_price
                        match = re.match(r"([<>]=?|=?)\s*(\d+(?:\.\d+)?)$", user_input)
                    elif column == 'theta':
                        # Matches decimal conditions for theta
                        match = re.match(r"([<>]=?|=?)\s*(\d+(?:\.\d+)?)$", user_input)
                    elif column == 'gamma':
                        # Matches decimal conditions for gamma
                        match = re.match(r"([<>]=?|=?)\s*(\d+(?:\.\d+)?)$", user_input)
                    elif column == 'open_interest_change':
                        # Matches numeric conditions which can be negative for open_interest_change
                        match = re.match(r"([<>]=?|=?)\s*(-?\d+)$", user_input)

                    if match:
                        operator, value = match.groups()
                        if column == 'expiry_date':
                            condition = f"{column} {operator} '{value}'"  # Dates need to be wrapped in quotes
                        else:
                            condition = f"{column} {operator} {value}"
                        conditions.append(condition)
                    else:
                        # If the input doesn't match the pattern, send an error message
                        await inter.response.send_message(f"Invalid format for {column}: {user_input}")
                        return
            # Combine all conditions with ' AND '
            condition_str = ' AND '.join(conditions)

            # Construct the query
            selected_fields = ', '.join(self.selected_columns)
            query = f"SELECT symbol, strike_price, call_put, expiry_date FROM options_data WHERE symbol = '{self.ticker}' AND {condition_str} LIMIT 25;"
            print(f"DEBUG QUERY: {query}")
            print(f"CONDITIONS STR DEBUG: {condition_str}")
            print(f"SELECTED FIELDS DEBUG: **{selected_fields}**")

            try:
                data = await options.fetch(query)
                if data:
                    df = pd.DataFrame(data, columns=['sym', 'strike', 'c/p', 'exp'])
                    table = tabulate(df, headers='keys', tablefmt='fancy', showindex=False)
                    embed = disnake.Embed(title="Database Query Results", description=f"```py\n{table}```", color=disnake.Colour.dark_orange())
                    embed.set_footer(text=f'Query: {query}')
                    await inter.response.send_message(embed=embed)
                else:
                    await inter.response.send_message("No data found for the given conditions.")
            except Exception as e:
                await inter.response.send_message(f"An error occurred: {e}")
            finally:
                await options.close_pool()






# Assuming WebullModal is already defined as shown in your previous message.

class VolumeAnalysisModal(disnake.ui.Modal):
    def __init__(self, ticker:str):
        self.ticker = ticker.upper()
        # Define placeholders for the volume_analysis table fields as needed
        placeholders = {
            'total_trades': "Enter total trades (e.g., '> 500')",
            'total_volume': "Enter total volume (e.g., '> 10000')",
            # Add more placeholders as needed
        }
        components = []
        for field, placeholder in placeholders.items():
            components.append(
                disnake.ui.TextInput(
                    label=field.capitalize().replace('_', ' '),
                    placeholder=placeholder,
                    custom_id=f"condition_{field}",
                    style=disnake.TextInputStyle.short,
                )
            )
        super().__init__(title="Filter Volume Analysis", components=components)

    async def callback(self, inter: disnake.ModalInteraction):
        # ... collect conditions for the volume_analysis table ...
        volume_conditions = []
        for component in self.components:
            user_input = inter.text_values[component.custom_id].strip()
            # Assuming user_input is in the form '<operator> <value>'
            match = re.match(r"([<>]=?|=?)\s*(\d+(?:\.\d+)?)$", user_input)
            if match:
                operator, value = match.groups()
                field = component.custom_id.replace('condition_', '')
                volume_conditions.append(f"{field} {operator} {value}")
            else:
                # If the input doesn't match the pattern, send an error message
                await inter.response.send_message(f"Invalid format for {field}: {user_input}")
                return

        # At this point, volume_conditions should contain all the conditions gathered from the modal.
        # You can now pass them to the function that will execute the JOIN query along with the conditions from the first modal.
        # For demonstration purposes, I'm just printing them here.
        print(volume_conditions)




# Assuming you have a class that handles state, like this:
class QueryState:
    def __init__(self):
        self.options_conditions = None
        self.volume_conditions = None




import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[2])
if project_dir not in sys.path:
    sys.path.append(project_dir)
from _markets.list_sets.ticker_lists import most_active_tickers
import os
from dotenv import load_dotenv
import asyncio
from asyncio import Lock
import asyncpg
import aiohttp
import pandas as pd
from .oic_models import OICOptionsMonitor
load_dotenv()
import requests
lock = Lock()
from datetime import datetime
import numpy as np

session = requests.session()
class AsyncOICSDK:
    def __init__(self, host, port, user, password, database):
        self.refresh_key_payload = {
            "clientKey": f"{os.environ.get('YOUR_OIC_KEY')}",
            "clientName": "OIC_test"
        }
        self.conn = None
        self.pool = None
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.token = None
        self.initialize_token()
        self.db_config = {
            "host": os.environ.get('DB_HOST', 'localhost'), # Default to this IP if 'DB_HOST' not found in environment variables
            "port": int(os.environ.get('DB_PORT')), # Default to 5432 if 'DB_PORT' not found
            "user": os.environ.get('DB_USER', 'postgres'), # Default to 'postgres' if 'DB_USER' not found
            "password": os.environ.get('DB_PASSWORD', 'fud'), # Use the password from environment variable or default
            "database": os.environ.get('DB_NAME', 'oic') # Database name for the new jawless database
        }
    def initialize_token(self):
        """
        Initialize the token by making the first request.
        """
        r = session.post(
            "https://ivlivefront.ivolatility.com/token/client/get",
            json=self.refresh_key_payload
        )
        if r.status_code == 200:
            self.token = r.text # Adjust the key according to actual response structure

    def build_headers(self):
        """
        Build headers using the current token.
        """
        if not self.token:
            self.initialize_token()
        return {'Authorization': f'Bearer {self.token}'}

    def refresh_token(self):
        """
        Refresh the token.
        """
        r = session.post(
            "https://ivlivefront.ivolatility.com/token/client/get",
            headers=self.build_headers(),
            json=self.refresh_key_payload
        )
        if r.status_code == 200:
            self.token = r
        return self.token
    async def get_option_id(self, ticker):
        """
        Gets the option ID using the lookup endpoint
        
        """
        url = f"https://private-authorization.ivolatility.com/lookup/?sortField=SYMBOL&symbol={ticker}&region=1&matchingType=EXACTLY&pageSize=10&page=0"
        async with aiohttp.ClientSession(headers=self.build_headers()) as session:
            async with session.get(url) as resp:
                try:
                    r = await resp.json()
                    print(r)
                    pages = r['page']
                    id = pages[0]['stockId']
                    return id
                except Exception as e:
                    print(e)
                    return None
    
    async def connect(self):
       
        self.pool = await asyncpg.create_pool(**self.db_config)

        return self.pool
    async def create_table(self, df, table_name, unique_column):
     
        print("Connected to the database.")
        dtype_mapping = {
            'int64': 'INTEGER',
            'float64': 'FLOAT',
            'object': 'TEXT',
            'bool': 'BOOLEAN',
            'datetime64': 'TIMESTAMP',
            'datetime64[ns]': 'timestamp',
            'datetime64[ms]': 'timestamp',
            'datetime64[ns, US/Eastern]': 'TIMESTAMP WITH TIME ZONE'
        }
        print(f"DataFrame dtypes: {df.dtypes}")
        # Check for large integers and update dtype_mapping accordingly
        for col, dtype in zip(df.columns, df.dtypes):
            if dtype == 'int64':
                max_val = df[col].max()
                min_val = df[col].min()
                if max_val > 2**31 - 1 or min_val < -2**31:
                    dtype_mapping['int64'] = 'BIGINT'
        history_table_name = f"{table_name}_history"
        async with self.pool.acquire() as connection:

            table_exists = await connection.fetchval(f"SELECT to_regclass('{table_name}')")
            
            if table_exists is None:
                unique_constraint = f'UNIQUE ({unique_column})' if unique_column else ''
                create_query = f"""
                CREATE TABLE {table_name} (
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))},
                    "insertion_timestamp" TIMESTAMP,
                    {unique_constraint}
                )
                """
                print(f"Creating table with query: {create_query}")

                # Create the history table
                history_create_query = f"""
                CREATE TABLE IF NOT EXISTS {history_table_name} (
                    id serial PRIMARY KEY,
                    operation CHAR(1) NOT NULL,
                    changed_at TIMESTAMP NOT NULL DEFAULT current_timestamp,
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))}
                );
                """
                print(f"Creating history table with query: {history_create_query}")
                await connection.execute(history_create_query)
                try:
                    await connection.execute(create_query)
                    print(f"Table {table_name} created successfully.")
                except asyncpg.UniqueViolationError as e:
                    print(f"Unique violation error: {e}")
            else:
                print(f"Table {table_name} already exists.")
            
            # Create the trigger function
            trigger_function_query = f"""
            CREATE OR REPLACE FUNCTION save_to_{history_table_name}()
            RETURNS TRIGGER AS $$
            BEGIN
                INSERT INTO {history_table_name} (operation, changed_at, {', '.join(f'"{col}"' for col in df.columns)})
                VALUES (
                    CASE
                        WHEN (TG_OP = 'DELETE') THEN 'D'
                        WHEN (TG_OP = 'UPDATE') THEN 'U'
                        ELSE 'I'
                    END,
                    current_timestamp,
                    {', '.join('OLD.' + f'"{col}"' for col in df.columns)}
                );
                RETURN NEW;
            END;
            $$ LANGUAGE plpgsql;
            """
            await connection.execute(trigger_function_query)

            # Create the trigger
            trigger_query = f"""
            DROP TRIGGER IF EXISTS tr_{history_table_name} ON {table_name};
            CREATE TRIGGER tr_{history_table_name}
            AFTER UPDATE OR DELETE ON {table_name}
            FOR EACH ROW EXECUTE FUNCTION save_to_{history_table_name}();
            """
            await connection.execute(trigger_query)


            # Alter existing table to add any missing columns
            for col, dtype in zip(df.columns, df.dtypes):
                alter_query = f"""
                DO $$
                BEGIN
                    BEGIN
                        ALTER TABLE {table_name} ADD COLUMN "{col}" {dtype_mapping[str(dtype)]};
                    EXCEPTION
                        WHEN duplicate_column THEN
                        NULL;
                    END;
                END $$;
                """
                await connection.execute(alter_query)

    async def batch_insert_dataframe(self, df, table_name, unique_columns, batch_size=250):
            """
            WORKS - Creates table - inserts data based on DTYPES.
            
            """
        
            async with lock:
                if not await self.table_exists(table_name):
                    await self.create_table(df, table_name, unique_columns)
                
                # Debug: Print DataFrame columns before modifications
                #print("Initial DataFrame columns:", df.columns.tolist())
                
                df = df.copy()
                df.dropna(inplace=True)
                df['insertion_timestamp'] = [datetime.now() for _ in range(len(df))]

                # Debug: Print DataFrame columns after modifications
                #print("Modified DataFrame columns:", df.columns.tolist())
                
                records = df.to_records(index=False)
                data = list(records)


                async with self.pool.acquire() as connection:
                    column_types = await connection.fetch(
                        f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'"
                    )
                    type_mapping = {col: next((item['data_type'] for item in column_types if item['column_name'] == col), None) for col in df.columns}

                    async with connection.transaction():
                        insert_query = f"""
                        INSERT INTO {table_name} ({', '.join(f'"{col}"' for col in df.columns)}) 
                        VALUES ({', '.join('$' + str(i) for i in range(1, len(df.columns) + 1))})
                        ON CONFLICT ({unique_columns})
                        DO UPDATE SET {', '.join(f'"{col}" = excluded."{col}"' for col in df.columns)}
                        """
                
                        batch_data = []
                        for record in data:
                            new_record = []
                            for col, val in zip(df.columns, record):
                    
                                pg_type = type_mapping[col]

                                if val is None:
                                    new_record.append(None)
                                elif pg_type == 'timestamp' and isinstance(val, np.datetime64):
                                    new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))

                
                                elif isinstance(val, datetime):
                                    new_record.append(pd.Timestamp(val).to_pydatetime())
                                elif pg_type in ['timestamp', 'timestamp without time zone', 'timestamp with time zone'] and isinstance(val, np.datetime64):
                                    new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))  # Modified line
                                elif pg_type in ['double precision', 'real'] and not isinstance(val, str):
                                    new_record.append(float(val))
                                elif isinstance(val, np.int64):  # Add this line to handle numpy.int64
                                    new_record.append(int(val))
                                elif pg_type == 'integer' and not isinstance(val, int):
                                    new_record.append(int(val))
                                else:
                                    new_record.append(val)
                        
                            batch_data.append(new_record)

                            if len(batch_data) == batch_size:
                                try:
                                    
                                
                                    await connection.executemany(insert_query, batch_data)
                                    batch_data.clear()
                                except Exception as e:
                                    print(f"An error occurred while inserting the record: {e}")
                                    await connection.execute('ROLLBACK')
                                    raise

                    if batch_data:  # Don't forget the last batch
        
                        try:

                            await connection.executemany(insert_query, batch_data)
                        except Exception as e:
                            print(f"An error occurred while inserting the record: {e}")
                            await connection.execute('ROLLBACK')
                            raise
    async def save_to_history(self, df, main_table_name, history_table_name):
        # Assume the DataFrame `df` contains the records to be archived
        if not await self.table_exists(history_table_name):
            await self.create_table(df, history_table_name, None)

        df['archived_at'] = datetime.now()  # Add an 'archived_at' timestamp
        await self.batch_insert_dataframe(df, history_table_name, None)
    async def table_exists(self, table_name):
        query = f"SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = '{table_name}');"
  
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                exists = await conn.fetchval(query)
        return exists


    async def most_active_options(self):
        """
        Gets the most active options from the Options Industry Council
        
        """
        url=f"https://private-authorization.ivolatility.com/favorites/instruments/most-active"
        data = await session.get(url, headers=self.build_headers()).json()


        print(data)

        df = pd.DataFrame(data)
        return df

    async def get_price(self, ticker):
        url = f"https://api.polygon.io/v2/last/nbbo/{ticker}?apiKey={os.environ.get('YOUR_POLYGON_KEY')}"
        async with aiohttp.ClientSession(headers=self.build_headers()) as session:
            async with session.get(url) as resp:
                resp = await resp.json()
                results = resp['results'] if 'results' in resp else None
                if results is not None:
                    price = results['P']
                    print(price)
                    return price

    async def options_monitor(self, ticker):
        stock_id = await self.get_option_id(ticker)
        if stock_id is not None:
            stock_price = await self.get_price(ticker)
            print(stock_price)
            if stock_price is not None:
            
                url = f"https://private-authorization.ivolatility.com/options-monitor/listOptionDataRow?stockId={stock_id}&center={stock_price}&regionId=1&strikesN=75&columns=alpha&columns=ask&columns=asksize&columns=asktime&columns=bid&columns=bidsize&columns=bidtime&columns=change&columns=changepercent&columns=delta&columns=theoprice&columns=gamma&columns=ivint&columns=ivask&columns=ivbid&columns=mean&columns=openinterest_eod&columns=optionsymbol&columns=volume&columns=paramvolapercent_eod&columns=alpha_eod&columns=ask_eod&columns=bid_eod&columns=delta_eod&columns=theoprice_eod&columns=gamma_eod&columns=ivint_eod&columns=mean_eod&columns=rho_eod&columns=theta_eod&columns=vega_eod&columns=changepercent_eod&columns=change_eod&columns=volume_eod&columns=quotetime&columns=rho&columns=strike&columns=style&columns=theta&columns=vega&columns=expirationdate&columns=forwardprice&columns=forwardprice_eod&columns=days&columns=days_eod&columns=iv&columns=iv_eod&rtMode=RT&userId=9999999&uuid=null"
                async with aiohttp.ClientSession(headers=self.build_headers()) as session:
                    async with session.get(url) as resp:
                        data = await resp.json()
                        print(data)
                        data = OICOptionsMonitor(resp)
                        if data is not None:
                            df = data.as_dataframe
                            await self.batch_insert_dataframe(df, table_name='monitor', unique_columns='put_optionsymbol, call_optionsymbol')

                    

                    return data
            


    async def monitor_all(self):
        tasks = [self.options_monitor(i) for i in most_active_tickers]


        await asyncio.gather(*tasks)
            







import pandas as pd
class CostDistribution:
    def __init__(self, r):
        all_data_dicts = []
        datas = r['data']
        for data in datas:
            self.tickerId = data['tickerId']
            self.avgCost = data['avgCost'] if 'avgCost' in data else None
            self.closeProfitRatio = data['closeProfitRatio'] if 'closeProfitRatio' in data else None
            self.chip90Start = data['chip90Start'] if 'chip90Start' in data else None
            self.chip90End = data['chip90End'] if 'chip90End' in data else None
            self.chip90Ratio = data['chip90Ratio'] if 'chip90Ratio' in data else None
            self.chip70Start = data['chip70Start'] if 'chip70Start' in data else None
            self.chip70End = data['chip70End'] if 'chip70end' in data else None
            self.chip70Ratio = data['chip70Ratio'] if 'chip70Ratio' in data else None
            self.close = data['close'] if 'close' in data else None
            self.totalShares = data['totalShares'] if 'totalShares' in data else None
            self.distributions = data['distributions'] if 'distributions' in data else None
            self.tradeStamp = data['tradeStamp'] if 'tradeStamp' in data else None


            self.data_dict = { 
                'Avg Cost': self.avgCost,
                '% Shareholders in Profit': self.closeProfitRatio,
                'Close': self.close,
                'Total Shares': self.totalShares,
                'Distributions': self.distributions,
                
            }
            all_data_dicts.append(self.data_dict)

        self.df = pd.DataFrame(all_data_dicts)


    def __str__(self):
        return f"CostDistribution(tickerId={self.tickerId}, avgCost={self.avgCost}, closeProfitRatio={self.closeProfitRatio}, chip90Start={self.chip90Start}, chip90End={self.chip90End}, chip90Ratio={self.chip90Ratio}, chip70Start={self.chip70Start}, chip70End={self.chip70End}, chip70Ratio={self.chip70Ratio}, close={self.close}, totalShares={self.totalShares}, distributions={self.distributions}, tradeStamp={self.tradeStamp})"

    def __repr__(self):
        return self.__str__()



from dataclasses import dataclass
from datetime import datetime

@dataclass
class BorrowedSharesData:
    ticker: str
    fee: float
    available: int
    date_updated: datetime

    def insert_into_pg(self, conn):
        cursor = conn.cursor()
        insert_query = """INSERT INTO borrowed_shares (ticker, fee, available, date_updated) VALUES (%s, %s, %s, %s);"""
        cursor.execute(insert_query, (self.ticker, self.fee, self.available, self.date_updated))
        conn.commit()



import disnake

from disnake.ext import commands

from discord_.bot_menus.pagination import AlertMenus
from apis.federal_register.fed_register_sdk import FedRegisterSDK
from apis.helpers import remove_html_tags
import pandas as pd

fedreg = FedRegisterSDK()


class FedRegisterCog(commands.Cog):
    """
    Fetch documents from the federal register!
    
    """


    def __init__(self, bot):
        self.bot=bot




    @commands.slash_command()
    async def register(self, inter):
        pass



    @register.sub_command()
    async def query_document(self, inter:disnake.AppCmdInter, query:str, effective_date_greater_than:str='2019-09-17', effective_date_less_than:str=None, order=commands.Param(choices=['newest','oldest','relevance']), is_significant=commands.Param(choices=['0', '1'])):
        """Provide input text to search for documents! Also can use optional data parameters!"""
        try:

            if effective_date_less_than is None:
                effective_date_less_than = fedreg.today


            await inter.response.defer(ephemeral=False)


            await fedreg.query_document(query,effective_date_greater_than=effective_date_greater_than, effective_date_less_than=effective_date_less_than,order=order, is_significant=is_significant)
            filename='data/fed_register/query_results.csv'
            df = pd.read_csv(filename)

            embeds = []
            for i,row in df.iterrows():
                result_count = row['result_count']
                search_description	= row['search_description']
                total_pages	= row['total_pages']
                title = row['title']
                type = row['type']
                abstract = row['abstract']
                document_number	= row['document_number']
                html_url = row['html_url']
                pdf_url	= row['pdf_url']
                public_inspection_url = row['public_inspection_url']
                publication_date = row['publication_date']
                excerpts = remove_html_tags(row['excerpts'])
                agency = row['agency']

                embed = disnake.Embed(title=f'Federal Register - {query}', description=f"```py\nDocument: {title}\nFROM: **{agency}**\n\nExcerpt:\n\n{excerpts}```", color=disnake.Colour.dark_gold(), url=html_url)
                embed.add_field(name=f"Doc Info:", value=f"> Result Count: **{result_count}**\n> Total Pages: **{total_pages}**\n> Type: **{type}**\n> Pub. Date: **{publication_date}**", inline=False)
                embed.add_field(name=f"Links:", value=f"> HTML: {html_url}\n\n> PDF: **{pdf_url}**\n\n> Pub. Inspection: **{public_inspection_url}**", inline=False)
                embed.add_field(name=f"Abstract:", value=f"> **{abstract[:1010]}**")
                embed.add_field(name="Search Description", value=f"> **{search_description}**")
                embed.set_footer(text=f'Doc #: {document_number} | Implemented by FUDSTOP')

                embeds.append(embed)

            
    




        
                await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds), file=disnake.File('data/fed_register/query_results.csv'))
        except Exception as e:
            await inter.send(f"# > {e}")
    


def setup(bot: commands.Bot):
    bot.add_cog(FedRegisterCog(bot))

    print(f'Federal Register Ready!')



from disnake.ext import commands
import disnake
from apis.fed_print.fedprint_sdk import FedPrint
import pandas as pd

fedprint = FedPrint()


class FedPrintCog(commands.Cog):
    """
    Discord commands related to fed-print API
    """

    def __init__(self, bot):
        self.bot=bot



    @commands.slash_command()
    async def fedprint(self, inter):
        pass


    @fedprint.sub_command()
    async def item_search(self, inter:disnake.AppCmdInter, year:str=commands.Param(choices=['year:2023', 'year:2022', 'year:2021', 'year:2020', 'year:2019', 'year:2018', 'year:2017', 'year:2016']), limit:str='25'):
        """Search for Fed Print items based on year"""
        await inter.response.defer()

        items = await fedprint.search(filter=year, limit=limit)
        df = items.as_dataframe
        filename='fudstop/data/fed_print/search_results.csv'
        df.to_csv(filename, index=False)
        await inter.edit_original_message(file=disnake.File(filename))


    @fedprint.sub_command()
    async def get_all(self, inter:disnake.AppCmdInter):
        """Get an abstract for a Fed Document"""
        await inter.response.defer()
        all_dataframes = []
        ids = await fedprint.get_series_id()
        for id in ids:
            series_info = await fedprint.get_series(id)
            all_dataframes.append(series_info.as_dataframe)

        final_dataframe = pd.concat(all_dataframes, ignore_index=True)

        filename = 'fudstop/data/fed_print/all_data.csv'

        final_dataframe.to_csv(filename, index=False)

        await inter.edit_original_message(file=disnake.File(filename))


    @fedprint.sub_command()
    async def get_urls(self, inter:disnake.AppCmdInter, limit:str):
        """Returns URLs only from fedprint"""
        await inter.response.defer()


        urls = await fedprint.print_file_urls(limit=limit)
        await inter.edit_original_message(', '.join(urls))




def setup(bot: commands.Bot):
    bot.add_cog(FedPrintCog(bot))
    print(f'Fed Print API Commands - Loaded!')


tools = [
    {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "format": {
                        "type": "string",
                        "enum": ["celsius", "fahrenheit"],
                        "description": "The temperature unit to use. Infer this from the users location.",
                    },
                },
                "required": ["location", "format"],
            },
        }
    },
    {
        "type": "function",
        "function": {
            "name": "get_n_day_weather_forecast",
            "description": "Get an N-day weather forecast",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "format": {
                        "type": "string",
                        "enum": ["celsius", "fahrenheit"],
                        "description": "The temperature unit to use. Infer this from the users location.",
                    },
                    "num_days": {
                        "type": "integer",
                        "description": "The number of days to forecast",
                    }
                },
                "required": ["location", "format", "num_days"]
            },
        }
    },
]